import json
import os
from typing import List, Optional, Tuple

import requests
from unify.exceptions import UnifyError

_available_dynamic_modes = [
    "lowest-input-cost",
    "input-cost",
    "lowest-output-cost",
    "output-cost",
    "lowest-itl",
    "itl",
    "lowest-ttft",
    "ttft",
    "highest-tks-per-sec",
    "tks-per-sec",
]

_base_url = "https://api.unify.ai/v0"


def _res_to_list(response: requests.Response) -> List[str]:
    return json.loads(response.text)


def list_models() -> List[str]:
    """
    Get a list of available models.

    Returns:
        List[str]: A list of available model names if successful, otherwise an empty list.
    Raises:
        BadRequestError: If there was an HTTP error.
        ValueError: If there was an error parsing the JSON response.
    """
    url = f"{_base_url}/models"
    return _res_to_list(requests.get(url))


def list_endpoints(model: str) -> List[str]:
    """
    Get a list of endpoints for a specific model.

    Args:
        model (str): The name of the model.

    Returns:
        List[str]: A list of endpoint names associated with the model if successful,
        otherwise an empty list.
    Raises:
        BadRequestError: If there was an HTTP error.
        ValueError: If there was an error parsing the JSON response.
    """
    url = f"{_base_url}/endpoints_of"
    return _res_to_list(requests.get(url, params={"model": model}))


def list_providers(model: str) -> List[str]:
    """
    Get a list of providers for a specific model.

    Args:
        model (str): The name of the model.

    Returns:
        List[str]: A list of provider names associated with the model if successful,
        otherwise an empty list.
    Raises:
        BadRequestError: If there was an HTTP error.
        ValueError: If there was an error parsing the JSON response.
    """
    url = f"{_base_url}/providers_of"
    return _res_to_list(requests.get(url, params={"model": model}))


def _validate_api_key(api_key: Optional[str]) -> str:
    if api_key is None:
        api_key = os.environ.get("UNIFY_KEY")
    if api_key is None:
        raise KeyError(
            "UNIFY_KEY is missing. Please make sure it is set correctly!",
        )
    return api_key


def _validate_endpoint_name(value: str) -> Tuple[str, str]:
    error_message = "endpoint string must use OpenAI API format: <uploaded_by>/<model_name>@<provider_name>"  # noqa: E501

    if not isinstance(value, str):
        raise UnifyError(error_message)

    try:
        model_name, provider_name = value.split("/")[-1].split("@")
    except ValueError:
        raise UnifyError(error_message)

    if not model_name or not provider_name:
        raise UnifyError(error_message)
    return (model_name, provider_name)


def _validate_endpoint(  # noqa: WPS231
    endpoint: Optional[str] = None,
    model: Optional[str] = None,
    provider: Optional[str] = None,
) -> Tuple[str, str, Optional[str]]:
    error_message = (
        "You must either provide an endpoint or the model and provider names!"
    )
    if endpoint:
        if model or provider:
            raise UnifyError(error_message)
        model, provider = _validate_endpoint_name(endpoint)  # noqa: WPS414
    else:
        if not model or not provider:
            raise UnifyError(error_message)
        endpoint = "@".join([model, provider])

    if provider in _available_dynamic_modes:
        provider = None
    return endpoint, model, provider



import sys
from typing import Dict, Generator, List, Optional

from unify.clients import Unify
from unify.exceptions import UnifyError


class ChatBot:  # noqa: WPS338
    """Agent class represents an LLM chat agent."""

    def __init__(
        self,
        endpoint: Optional[str] = None,
        model: Optional[str] = None,
        provider: Optional[str] = None,
        api_key: Optional[str] = None,
    ) -> None:
        """
        Initializes the ChatBot object.

        Args:
            endpoint (str, optional): Endpoint name in OpenAI API format:
                <uploaded_by>/<model_name>@<provider_name>
                Defaults to None.

            model (str, optional): Name of the model. If None,
            endpoint must be provided.

            provider (str, optional): Name of the provider. If None,
            endpoint must be provided.

            api_key (str, optional): API key for accessing the Unify API.
                If None, it attempts to retrieve the API key from the
                environment variable UNIFY_KEY.
                Defaults to None.

        Raises:
            UnifyError: If the API key is missing.
        """
        self._message_history: List[Dict[str, str]] = []
        self._paused = False
        self._client = Unify(
            api_key=api_key,
            endpoint=endpoint,
            model=model,
            provider=provider,
        )

    @property
    def client(self) -> Unify:
        """
        Get the client object.  # noqa: DAR201.

        Returns:
            str: The model name.
        """
        return self._client

    def set_client(self, value: Unify) -> None:
        """
        Set the model name.  # noqa: DAR101.

        Args:
            value: The unify client.
        """
        if isinstance(value, Unify):
            self._client = value
        else:
            raise UnifyError("Invalid client!")

    @property
    def model(self) -> str:
        """
        Get the model name.  # noqa: DAR201.

        Returns:
            str: The model name.
        """
        return self._client.model

    def set_model(self, value: str) -> None:
        """
        Set the model name.  # noqa: DAR101.

        Args:
            value (str): The model name.
        """
        self._client.set_model(value)
        if self._client.provider:
            self._client.set_endpoint("@".join([value, self._client.provider]))
        else:
            mode = self._client.endpoint.split("@")[1]
            self._client.set_endpoint("@".join([value, mode]))

    @property
    def provider(self) -> Optional[str]:
        """
        Get the provider name.  # noqa :DAR201.

        Returns:
            str: The provider name.
        """
        return self._client.provider

    def set_provider(self, value: str) -> None:
        """
        Set the provider name.  # noqa: DAR101.

        Args:
            value (str): The provider name.
        """
        self._client.set_provider(value)
        self._client.set_endpoint("@".join([self._client._model, value]))

    @property
    def endpoint(self) -> str:
        """
        Get the endpoint name.  # noqa: DAR201.

        Returns:
            str: The endpoint name.
        """
        return self._client.endpoint

    def set_endpoint(self, value: str) -> None:
        """
        Set the model name.  # noqa: DAR101.

        Args:
            value (str): The endpoint name.
        """
        self._client.set_endpoint(value)
        self._client.set_model(value.split("@")[0])
        self._client.set_provider(value.split("@")[1])

    def _get_credits(self) -> float:
        """
        Retrieves the current credit balance from associated with the UNIFY account.

        Returns:
            float: Current credit balance.
        """
        return self._client.get_credit_balance()

    def _process_input(
        self, inp: str, show_credits: bool, show_provider: bool
    ) -> Generator[str, None, None]:
        """
        Processes the user input to generate AI response.

        Args:
            inp (str): User input message.
            show_credits (bool): Whether to show credit consumption.
            show_credits (bool): Whether to show provider used.

        Yields:
            str: Generated AI response chunks.
        """
        self._update_message_history(role="user", content=inp)
        initial_credit_balance = self._get_credits()
        stream = self._client.generate(
            messages=self._message_history,
            stream=True,
        )
        words = ""
        for chunk in stream:
            words += chunk
            yield chunk

        self._update_message_history(
            role="assistant",
            content=words,
        )
        final_credit_balance = self._get_credits()
        if show_credits:
            sys.stdout.write(
                "\n(spent {:.6f} credits)".format(
                    initial_credit_balance - final_credit_balance,
                ),
            )
        if show_provider:
            sys.stdout.write("\n(provider: {})".format(self._client.provider))

    def _update_message_history(self, role: str, content: str) -> None:
        """
        Updates message history with user input.

        Args:
            role (str): Either "assistant" or "user".
            content (str): User input message.
        """
        self._message_history.append(
            {
                "role": role,
                "content": content,
            },
        )

    def clear_chat_history(self) -> None:
        """Clears the chat history."""
        self._message_history.clear()

    def run(self, show_credits: bool = False, show_provider: bool = False) -> None:
        """
        Starts the chat interaction loop.

        Args:
            show_credits (bool, optional): Whether to show credit consumption.
            Defaults to False.
            show_provider (bool, optional): Whether to show the provider used.
            Defaults to False.
        """
        if not self._paused:
            sys.stdout.write(
                "Let's have a chat. (Enter `pause` to pause and `quit` to exit)\n",
            )
            self.clear_chat_history()
        else:
            sys.stdout.write(
                "Welcome back! (Remember, enter `pause` to pause and `quit` to exit)\n",
            )
        self._paused = False
        while True:
            sys.stdout.write("> ")
            inp = input()
            if inp == "quit":
                self.clear_chat_history()
                break
            elif inp == "pause":
                self._paused = True
                break
            for word in self._process_input(inp, show_credits, show_provider):
                sys.stdout.write(word)
                sys.stdout.flush()
            sys.stdout.write("\n")
